#include <Python.h> 
#include <numpy/npy_common.h>
#include <numpy/ndarrayobject.h>
#include "schrodinger2d.h"


#define ARG_ERROR(e) if (e) {PyErr_SetString(PyExc_ValueError,"bad arguments"); return NULL;}

#define SCHRODINGER2D_ERROR(e) {\
    if (e<0) {\
        PyErr_SetString(PyExc_ValueError,"Erreur"); return NULL;\
    }\
}

static PyObject* py_schrodinger2d_open(PyObject *self, PyObject *args) {
    char type;
    int px_min,py_min,levels;
    float Lx;
    int normalize,colormap;
    float gamma;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiifiif",&type,&px_min,&py_min,&levels,&Lx,&normalize,&colormap,&gamma))
    SCHRODINGER2D_ERROR(schrodinger2d_open(type,px_min,py_min,levels,Lx,normalize,colormap,gamma))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_close(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_close())
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_time_step(PyObject *self, PyObject *args) {
    float dt;
    ARG_ERROR(!PyArg_ParseTuple(args,"f",&dt))
    SCHRODINGER2D_ERROR(schrodinger2d_time_step(dt))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_potential_rectangle(PyObject *self, PyObject *args) {
    int ci0,cj0,width,height;
    float v;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiif",&ci0,&cj0,&width,&height,&v))
    SCHRODINGER2D_ERROR(schrodinger2d_potential_rectangle(ci0,cj0,width,height,v))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_potential_disk(PyObject *self, PyObject *args) {
    int ci0,cj0,radius;
    float v;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiif",&ci0,&cj0,&radius,&v))
    SCHRODINGER2D_ERROR(schrodinger2d_potential_disk(ci0,cj0,radius,v))
    Py_RETURN_NONE;
}


int NumpyArray2Array(PyArrayObject *np_array, float **array, int *size_x, int *size_y) {
    int type,size,k;
    float *data;
    npy_intp i,j;
    type = PyArray_TYPE(np_array);
    if (type!=NPY_FLOAT) return 0;
    if (PyArray_NDIM(np_array)!=2) return 0;
    *size_y = PyArray_DIM(np_array,0);
    *size_x = PyArray_DIM(np_array,1);
    size = (*size_x)*(*size_y);
    *array = (float*)malloc(sizeof(float)*(size));
    k = 0;
    for (j=0; j<*size_y; j++) {
        for (i=0; i<*size_x; i++) {
             data = PyArray_GETPTR2(np_array,j,i);
                 (*array)[k] = *data;
                 k++;
             }
        }
    return 1;
}

static PyObject* py_schrodinger2d_potential(PyObject *self, PyObject *args) {
    PyObject *V_array;
    float *data;
    int size_x,size_y;
    
    ARG_ERROR(!PyArg_ParseTuple(args,"O",&V_array))
    if( NumpyArray2Array((PyArrayObject*)V_array,&data,&size_x,&size_y)==0) {
           
    }
    
    SCHRODINGER2D_ERROR(schrodinger2d_potential(data))
    free(data);
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_schrodinger(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_schrodinger())
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_dirichlet_borders(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_dirichlet_borders())
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_dirichlet_line(PyObject *self, PyObject *args) {
    int i0,j0,length,direction;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiii",&i0,&j0,&length,&direction))
    SCHRODINGER2D_ERROR(schrodinger2d_dirichlet_line(i0,j0,length,direction))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_init(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_init())
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_iterations(PyObject *self, PyObject *args) {
    float ti,tf;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&ti,&tf))
    SCHRODINGER2D_ERROR(schrodinger2d_iterations(ti,tf))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_iterations_2threads(PyObject *self, PyObject *args) {
    float ti,tf;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&ti,&tf))
    SCHRODINGER2D_ERROR(schrodinger2d_iterations_2threads(ti,tf))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_iterations_4threads(PyObject *self, PyObject *args) {
    float ti,tf;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&ti,&tf))
    SCHRODINGER2D_ERROR(schrodinger2d_iterations_4threads(ti,tf))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_paquet(PyObject *self, PyObject *args) {
    float x0,y0,k0,sigma0;
    ARG_ERROR(!PyArg_ParseTuple(args,"ffff",&x0,&y0,&k0,&sigma0))
    SCHRODINGER2D_ERROR(schrodinger2d_paquet(x0,y0,k0,sigma0))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_get_psi(PyObject *self, PyObject *args) {
    int width,height;
    npy_intp dims[2], c[2];
    PyObject *numpy_array_re, *numpy_array_im;
    void *data_re,*data_im;
    schrodinger2d_get_size(&width,&height);
    dims[0] = height;
    dims[1] = width;
    c[0] = 0; c[1] = 0;
    numpy_array_re = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    numpy_array_im = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data_re = PyArray_GetPtr((PyArrayObject*)numpy_array_re,c);
    data_im = PyArray_GetPtr((PyArrayObject*)numpy_array_im,c);
    SCHRODINGER2D_ERROR(schrodinger2d_get_psi(data_re,data_im))
    return Py_BuildValue("OO",numpy_array_re,numpy_array_im); 
}

static PyObject* py_schrodinger2d_get_proba(PyObject *self, PyObject *args) {
    int width,height;
    npy_intp dims[2], c[2];
    PyObject *numpy_array;
    void *data;
    schrodinger2d_get_size(&width,&height);
    dims[0] = height;
    dims[1] = width;
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    SCHRODINGER2D_ERROR(schrodinger2d_get_proba(data))
    return Py_BuildValue("O",numpy_array); 
}



static PyObject* py_schrodinger2d_get_proba_colors(PyObject *self, PyObject *args) {
    int width,height;
    npy_intp dims[3], c[3];
    PyObject *numpy_array;
    void *data;
    schrodinger2d_get_size(&width,&height);
    dims[0] = height;
    dims[1] = width;
    dims[2] = 4;
    c[0] = 0; c[1] = 0; c[2] = 0;
    numpy_array = PyArray_SimpleNew(3,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    SCHRODINGER2D_ERROR(schrodinger2d_get_proba_colors(data))
    return Py_BuildValue("O",numpy_array);
}

static PyObject* py_schrodinger2d_get_contours(PyObject *self, PyObject *args) {
    int width,height;
    npy_intp dims[2], c[2];
    PyObject *numpy_array;
    void *data;
    schrodinger2d_get_size(&width,&height);
    dims[0] = height;
    dims[1] = width;
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    SCHRODINGER2D_ERROR(schrodinger2d_get_contours(data))
    return Py_BuildValue("O",numpy_array); 
}

/*---------------------------- openGL ---------------------------------------*/

static PyObject* py_schrodinger2d_start_gl(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_start_gl())
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_start_rendering(PyObject *self, PyObject *args) {
    float ti,tf;
    int opencl,rendering;
    float surface_height;
    float angle_x,angle_z;
    int width,height;
    ARG_ERROR(!PyArg_ParseTuple(args,"ffiiififf",&ti,&tf,&opencl,&width,&height,&surface_height,&rendering,&angle_x,&angle_z))
    SCHRODINGER2D_ERROR(schrodinger2d_start_rendering(ti,tf,opencl,width,height,surface_height,rendering,angle_x,angle_z))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_offscreen_rendering(PyObject *self, PyObject *args) {
    int rendering;
    float ti,tf;
    int opencl;
    float surface_height;
    int width,height;
    float angle_x,angle_z;
    npy_intp dims[3], c[3];
    PyObject *numpy_array;
    void *data;
    ARG_ERROR(!PyArg_ParseTuple(args,"ffiiififf",&ti,&tf,&opencl,&width,&height,&surface_height,&rendering,&angle_x,&angle_z))
    dims[0] = height;
    dims[1] = width;
    dims[2] = 3;
    c[0] = 0; c[1] = 0; c[2] = 0;
    numpy_array = PyArray_SimpleNew(3,dims,NPY_UINT8);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    SCHRODINGER2D_ERROR(schrodinger2d_offscreen_rendering(ti,tf,opencl,width,height,surface_height,rendering,angle_x,angle_z,data))
    
    return Py_BuildValue("O",numpy_array); 
}

/*---------------------------- openCL ---------------------------------------*/
#ifdef OPENCL
static PyObject* py_schrodinger2d_opencl_platforms(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_opencl_platforms())
    Py_RETURN_NONE;  
}

static PyObject* py_schrodinger2d_set_opencl_platform_device(PyObject *self, PyObject *args) {
    int p,d;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&p,&d))
    SCHRODINGER2D_ERROR(schrodinger2d_set_opencl_platform_device(p,d))
    Py_RETURN_NONE;
}

static PyObject* py_schrodinger2d_opencl_init(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_opencl_init())
    Py_RETURN_NONE;  
}

static PyObject* py_schrodinger2d_opencl_create_memory(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_opencl_create_memory())
    Py_RETURN_NONE;  
}

static PyObject* py_schrodinger2d_opencl_release_memory(PyObject *self, PyObject *args) {
    SCHRODINGER2D_ERROR(schrodinger2d_opencl_release_memory())
    Py_RETURN_NONE;  
}

static PyObject* py_schrodinger2d_opencl_iterations(PyObject *self, PyObject *args) {
    float ti,tf;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&ti,&tf))
    SCHRODINGER2D_ERROR(schrodinger2d_opencl_iterations(ti,tf))
    Py_RETURN_NONE;
}

#endif



static PyMethodDef pyschrodinger2dMethods[] = {
    {"open",py_schrodinger2d_open,METH_VARARGS},
    {"close",py_schrodinger2d_close,METH_NOARGS},
    {"time_step",py_schrodinger2d_time_step,METH_VARARGS},
    {"potential_rectangle",py_schrodinger2d_potential_rectangle,METH_VARARGS},
    {"potential_disk",py_schrodinger2d_potential_disk,METH_VARARGS},
    {"potential",py_schrodinger2d_potential,METH_VARARGS},
    {"schrodinger",py_schrodinger2d_schrodinger,METH_NOARGS},
    {"dirichlet_borders",py_schrodinger2d_dirichlet_borders,METH_VARARGS},
    {"dirichlet_line",py_schrodinger2d_dirichlet_line,METH_VARARGS},
    {"init",py_schrodinger2d_init,METH_NOARGS},
    {"iterations",py_schrodinger2d_iterations,METH_VARARGS},
    {"iterations_2threads",py_schrodinger2d_iterations_2threads,METH_VARARGS},
    {"iterations_4threads",py_schrodinger2d_iterations_4threads,METH_VARARGS},
    {"paquet",py_schrodinger2d_paquet,METH_VARARGS},
    {"get_psi",py_schrodinger2d_get_psi,METH_NOARGS},
    {"get_proba",py_schrodinger2d_get_proba,METH_NOARGS},
    {"get_proba_colors",py_schrodinger2d_get_proba_colors,METH_NOARGS},
    {"get_contours",py_schrodinger2d_get_contours,METH_NOARGS},
    {"start_rendering",py_schrodinger2d_start_rendering,METH_VARARGS},
    {"offscreen_rendering",py_schrodinger2d_offscreen_rendering,METH_VARARGS},
    {"start_gl",py_schrodinger2d_start_gl,METH_NOARGS},
#ifdef OPENCL 
    {"opencl_platforms",py_schrodinger2d_opencl_platforms,METH_NOARGS},
    {"set_opencl_platform_device",py_schrodinger2d_set_opencl_platform_device,METH_VARARGS},
    {"opencl_init",py_schrodinger2d_opencl_init,METH_NOARGS},
    {"opencl_create_memory",py_schrodinger2d_opencl_create_memory,METH_NOARGS},
    {"opencl_release_memory",py_schrodinger2d_opencl_release_memory,METH_NOARGS},
    {"opencl_iterations",py_schrodinger2d_opencl_iterations,METH_VARARGS},
#endif
    {NULL,NULL}
};

#ifdef PYTHON3
static struct PyModuleDef pyschrodinger2dmodule = {
    PyModuleDef_HEAD_INIT,
    "pyschrodinger2d", 
    0,
    -1,
    pyschrodinger2dMethods
};

PyMODINIT_FUNC
PyInit_pyschrodinger2d(void) {
    import_array();
    return PyModule_Create(&pyschrodinger2dmodule);
    
}
#else
void initpyschrodinger2d(void) {
    Py_InitModule("pyschrodinger2d",pyschrodinger2dMethods);
    import_array();
}
#endif
